# Copyright (c) 2011-2012 Cloudera, Inc. All rights reserved.

import logging
import os
import threading
import time

from cmf.monitor import constants
from cmf.monitor.count_down_latch import CountDownLatch
from cmf.monitor.datanode import DataNodeMonitor
from cmf.monitor.firehose import Firehose
from cmf.monitor.host import HostMonitor
from cmf.monitor.namenode import NameNodeMonitor
from cmf.monitor.secondarynamenode import SecondaryNameNodeMonitor
from cmf.monitor.jobtracker import JobTrackerMonitor
from cmf.monitor.tasktracker import TaskTrackerMonitor, init_tasktracker_monitor
from cmf.monitor.master import MasterMonitor
from cmf.monitor.regionserver import RegionServerMonitor
from cmf.monitor.servicemonitor import ServiceMonitorMonitor
from cmf.monitor.activitymonitor import ActivityMonitorMonitor
from cmf.monitor.eventserver import EventServerMonitor
from cmf.monitor.alertpublisher import AlertPublisherMonitor
from cmf.monitor.reportsmanager import ReportsManagerMonitor
from cmf.monitor.hostmonitor import HostMonitorMonitor
from cmf.monitor.journalnode import JournalNodeMonitor
from cmf.monitor.zookeeper import ZooKeeperServerMonitor
from cmf.monitor.flume import FlumeAgentMonitor
from cmf.monitor.impalad import ImpaladMonitor
from cmf.monitor.statestore import StateStoreMonitor
from cmf.monitor.host_update import HostUpdate
from cmf.monitor.registry import MonitorRegistry
from cmf.monitor.wakeable_thread import WakeableThread
from cmf.monitor.host.filesystem_map import FilesystemMap

LOG = logging.getLogger("monitor")
_EMPTY_HOST_UPDATE = HostUpdate()

class MonitorDaemon(object):
  """
  The MonitorDaemon periodically report metrics to the firehose.
  """
  def __init__(self, hostname, host_id, filesystem_map=None, agent=None):
    self._agent = agent
    self._hostname = hostname
    self._host_id = host_id
    self._registry = MonitorRegistry()

    # This is set by the agent while it processes heartbeats, so folks should
    # take care while accessing it.
    self.enabled_metric_reporters = []

    # This thread is responsible for sending reports to the various firehoses.
    # It is awoken by the scheduling thread on a regular schedule (to send full
    # reports) and at special times when new monitors have been created and are
    # ready to send their first report (partial reports).
    self._reporting_thread = \
        WakeableThread("MonitorDaemon-Reporter", self._report)
    # This thread is responsible for waking up the reporitng thread when it is
    # appropriate. It also drives detecting when new monitors are ready to send
    # their first report.
    self._scheduling_thread = \
        WakeableThread("MonitorDaemon-Scheduler", self._schedule)

    # This holds our reporting interval, which is specified in seconds. This
    # comes from the SCM heartbeat, so the initial value doesn't matter a whole
    # lot.
    self._reporting_interval_seconds = 5.0

    # This is the time at which we last woke the reporter thread to have it send
    # a full report.
    self._last_full_report_scheduled = None
    # This flag is set by the scheduling thread to indicate that the reporting
    # thread should send a full report when it next runs. It is cleared by the
    # reporting thread.
    self._full_report = False

    # This lock is used to protect access to the various new monitor members
    self._new_monitors_lock = threading.Lock()
    # This is a map from monitor object to 'expiration time'. A monitor is put
    # into this map when it is first created and it stays in the map until it
    # has either indicated that it is ready to send reports or the expiration
    # time has passed.
    self._new_monitors = {}
    # When a new monitor is ready to send a report, it is moved from the map
    # above to this list.
    self._new_monitors_ready = []

    # We do not need to protect the _firehoses since they
    # are only accessed from this thread.
    # Map of role name -> Firehose object.
    self._firehoses = dict()
    # Array of dictionaries, coming from the heartbeat.
    # self._firehoses is updated by the MonitorDaemon
    # thread at an interface.  _update_firehose_connections()
    # updates self._firehoses based on self._firehoses_conn_info.
    self._firehoses_conn_info = []

    if filesystem_map is None:
      self._fs_map = FilesystemMap()
    else:
      self._fs_map = filesystem_map

    init_tasktracker_monitor()
    self._host_monitor = HostMonitor(self, self._registry)
    self.last_amon_message = None
    self.last_smon_message = None
    self.last_hmon_message = None

    if agent is None:
      # The agent is null during unit test runs, the timeouts don't matter in
      # that case, so just use whatever.
      self._metrics_url_timeout_seconds = 30.0
      self._task_metrics_timeout_seconds = 10.0
    else:
      self._metrics_url_timeout_seconds = agent.metrics_url_timeout_seconds
      self._task_metrics_timeout_seconds = agent.task_metrics_timeout_seconds

  def get_agent(self):
    return self._agent

  def start(self):
    self._host_monitor.ensure_collecting()
    self._reporting_thread.start()
    # We force the scheduling thread to run immediately.
    self._scheduling_thread.wake(CountDownLatch(1))
    self._scheduling_thread.start()

  def should_stop(self):
    self._host_monitor.stop()
    self._reporting_thread.should_stop()
    self._scheduling_thread.should_stop()

  def set_interval(self, seconds):
    self._reporting_interval_seconds = seconds

  def get_interval(self):
    return self._reporting_interval_seconds

  def get_filesystem_map(self):
    return self._fs_map

  def record_heartbeat_timestamps(self,
                                  ts_agent_send,
                                  ts_agent_recv,
                                  ts_master_send,
                                  ts_master_recv):

    self._host_monitor.record_heartbeat_timestamps(
        ts_agent_send, ts_agent_recv,
        ts_master_send, ts_master_recv)

  def update_host_config(self, host_config):
    """
    Called by agent after a heartbeat to update host config passed from SCM.
    """
    self._host_monitor.update_host_config(host_config)

  def update_firehoses(self, firehoses):
    """
    Called by agent after a heartbeat.

    firehoses is expected to be an array of dictionaries, with
    each dictionary having "rolename", "roletype", "address", "port", 
    "report_interval" fields.  For example:
    {u'roletype': u'FIREHOSE_ACTIVITY_MONITOR', u'rolename': u'm-FIREHOSE-2', u'port': 9999, u'report_interval': 10, u'address': u'air'}
    """
    self._firehoses_conn_info = firehoses

  def _update_firehose_connections(self):
    """
    Called to update the monitor's firehose connection information.
    Resolves the potential differences between self._firehoses and self._firehoses_connection_info
    """
    # Grab the conn_info reference, in case it gets updated
    # while we're here.  conn_info is not mutated by anyone;
    # it's always a fresh object coming from the heartbeat.
    conn_info = self._firehoses_conn_info

    # Local alias for self._firehoses
    firehoses = self._firehoses

    # rolename -> dict
    by_name = dict()
    for fh in conn_info:
      by_name[fh["rolename"]] = fh

    # Remove firehoses that don't exist any more
    to_remove = []
    for fh in firehoses.values():
      if fh.name not in by_name:
        to_remove.append(fh.name)
      else:
        # Let's update the connection
        # info.
        f = by_name[fh.name]
        fh.set_connection_info(f["address"], f["port"])

    for r in to_remove:
      LOG.info("Removing firehose: " + r)
      del firehoses[r]

    # Insert new ones
    for fh in conn_info:
      if fh["rolename"] not in firehoses:
        firehoses[fh["rolename"]] = \
            Firehose(fh["rolename"], fh["roletype"], fh["address"], fh["port"])
        LOG.info("Adding firehose: " + str(fh));
      
    # Minimal report interval
    if len(conn_info) == 0:
      pass
    else:
      interval = min([fh["report_interval"] for fh in conn_info])
      if interval != self.get_interval():
        LOG.info("Updating reporting interval: %s -> %s" 
          % (self.get_interval(), interval))
        self.set_interval(interval)

  def _report(self):
    # Whether we are doing a full report or a partial report, we can clear the
    # list of new montitors that are ready.
    self._new_monitors_lock.acquire()
    new_monitors_ready = self._new_monitors_ready
    self._new_monitors_ready = []
    self._new_monitors_lock.release()

    if self._full_report:
      self._full_report = False
      monitors = self._registry.get_registry()
    else:
      monitors = new_monitors_ready

    # Filter out the monitors which have metric reporting disabled. Reporting
    # might be disabled for some monitors if say the customer lacks a particular
    # license.
    enabled = self.enabled_metric_reporters
    monitors = filter(lambda x: x.get_subject_type() in enabled, monitors)

    if len(monitors) > 0:
      self._report_for_monitors(monitors)

  def _report_for_monitors(self, participants):
    """Collect reports from the Monitors and send them to the firehoses"""
    self._update_firehose_connections()

    # Update the process information of each of our monitors if necessary. Note
    # that if we are using supervisord events, then there's no need to do this
    # as any change in process status will have been communicated via an event
    # and the monitors will be updated by that event handling.
    if not self._agent.monitoring_supervisord_events():
      self._agent.update_monitor_process_information()

    # We signal each monitor in turn to start collecting then wait for each to
    # signal its reports are ready. We will wait at most the max_collection_wait
    # of the agent.
    latch = CountDownLatch(len(participants))
    for monitor in participants:
      monitor.start_collecting(latch)
    latch.await(self._agent.max_collection_wait_seconds)

    host_update = None
    role_updates = [ ]
    attempt_updates = [ ]
    service_updates = [ ]

    for monitor in participants:
      # We build a single large list of attempt updates. Each monitor can add to
      # it, but we store their updates in a single flat list.
      attempts = monitor.get_attempt_updates()
      if len(attempts) > 0:
        attempt_updates.extend(attempts)
        LOG.debug("%s reports attempt metrics: %s" % (monitor, attempts))

      # Each monitor gets to send one or more role updates.
      role_update = monitor.get_role_update()
      if role_update is not None:
        LOG.debug("%s reports role metrics: %s" % (monitor, role_update))
        role_updates.append(role_update.to_avro_dict())

      # Each monitor gets to send a service update.
      service_update = monitor.get_service_update()
      if service_update is not None:
        LOG.debug("%s reports service metrics: %s" % (monitor, service_update))
        service_updates.append(service_update.to_avro_dict())

      if isinstance(monitor, HostMonitor):
        host_update = self._host_monitor.get_host_update()
        LOG.debug("%s reports host metrics: %s" % (monitor, host_update))

    if len(self._firehoses.values()) == 0:
      LOG.debug("Not sending report since no firehoses are configured.")
      return

    for fh in self._firehoses.values():
      if fh.roletype == "ACTIVITYMONITOR":
        if len(attempt_updates) == 0:
          continue
        self.last_amon_message = \
            self._send_fh_message(fh,
                                  attempt_metrics=attempt_updates)
      elif fh.roletype == "SERVICEMONITOR":
        if len(service_updates) == 0 and len(role_updates) == 0:
          continue
        self.last_smon_message = \
            self._send_fh_message(fh,
                                  service_updates=service_updates,
                                  role_metrics=role_updates)
      elif fh.roletype == "HOSTMONITOR":
        if host_update is None:
          continue
        self.last_hmon_message = \
            self._send_fh_message(fh,
                                  host_update=host_update)

  def _send_fh_message(self,
                       firehose,
                       attempt_metrics = [],
                       host_update = _EMPTY_HOST_UPDATE,
                       service_updates = [],
                       role_metrics = []):
    agentmsg = dict(ts_secs=long(time.time()),
                    hostname=self._hostname,
                    attempt_metrics=attempt_metrics,
                    host_update=host_update.to_avro_dict(),
                    service_updates=service_updates,
                    role_metrics=role_metrics,
                    host_id=self._host_id,
                    version=constants.AGENT_VERSION)
    # Synthesize a 'agent_messages' Avro record to include the
    # agentmsg.
    agentmsgs = [ ]
    agentmsgs.append(agentmsg)
    agent_messages = dict(agent_msgs=agentmsgs)
    firehose.send(agent_messages)

    return agentmsg

  def make_monitor(self, process_name):
    """
    make_monitor(process_name) -> MonitorInterface

    The returned monitor is not running yet. The caller needs to
    refresh() it to activate it.

    Returns None if monitoring is not supported.
    """
    # Find out which type it is
    if process_name == 'hdfs-NAMENODE':
      return NameNodeMonitor(self, self._registry)
    elif process_name == 'hdfs-SECONDARYNAMENODE':
      return SecondaryNameNodeMonitor(self, self._registry)
    elif process_name == 'hdfs-DATANODE':
      return DataNodeMonitor(self, self._registry)
    elif process_name == 'mapreduce-JOBTRACKER':
      return JobTrackerMonitor(self, self._registry)
    elif process_name == 'mapreduce-TASKTRACKER':
      return TaskTrackerMonitor(self, self._registry)
    elif process_name == 'hbase-REGIONSERVER':
      return RegionServerMonitor(self, self._registry)
    elif process_name == 'hbase-MASTER':
      return MasterMonitor(self, self._registry)
    elif process_name == 'zookeeper-server':
      return ZooKeeperServerMonitor(self, self._registry)
    elif process_name == 'flume-AGENT':
      return FlumeAgentMonitor(self, self._registry)
    elif process_name == 'cloudera-mgmt-SERVICEMONITOR':
      return ServiceMonitorMonitor(self, self._registry)
    elif process_name == 'cloudera-mgmt-ACTIVITYMONITOR':
      return ActivityMonitorMonitor(self, self._registry)
    elif process_name == 'cloudera-mgmt-EVENTSERVER':
      return EventServerMonitor(self, self._registry)
    elif process_name == 'cloudera-mgmt-ALERTPUBLISHER':
      return AlertPublisherMonitor(self, self._registry)
    elif process_name == 'cloudera-mgmt-REPORTSMANAGER':
      return ReportsManagerMonitor(self, self._registry)
    elif process_name == 'cloudera-mgmt-HOSTMONITOR':
      return HostMonitorMonitor(self, self._registry)
    elif process_name == 'impala-IMPALAD':
      return ImpaladMonitor(self, self._registry)
    elif process_name == 'impala-STATESTORE':
      return StateStoreMonitor(self, self._registry)
    elif process_name == 'hdfs-JOURNALNODE':
      return JournalNodeMonitor(self, self._registry)

    return None

  def get_registry_for_testing(self):
    return self._registry

  def get_firehose_debug_status_reports(self):
    """
    Returns map of name -> debug status reports for every firehose.
    """
    ret = {}
    for k, fh in self._firehoses.items():
      ret[k] = fh.get_debug_status_report()
    return ret

  def get_metrics_url_timeout_seconds(self):
    return self._metrics_url_timeout_seconds

  def get_task_metrics_timeout_seconds(self):
    return self._task_metrics_timeout_seconds

  def _schedule(self):
    # Check if we need to schedule a full report, and determine the starting
    # point for our sleep time.
    sleep_time = self._time_until_next_full_report()
    if sleep_time <= 0.0:
      self._full_report = True
      sleep_time = self._reporting_interval_seconds

    # If a new monitor has just become ready, we may need to send a partial
    # report (this will be ignored if we're already going to send a full
    # report). At the same time, we check the number of new monitors that
    # we will have after such a report is sent. If there are any, then we will
    # shorten our sleep, so we can quickly detect if those monitors become
    # ready. This call also clears any expired monitors from the map of
    # new monitors.
    num_monitors_ready, num_new_monitors = self._update_new_monitors()
    if num_new_monitors > 0:
      sleep_time = 1.0

    if self._full_report:
      self._reporting_thread.wake(CountDownLatch(1))
      self._last_full_report_scheduled = time.time()
    elif num_monitors_ready > 0:
      self._reporting_thread.wake(CountDownLatch(1))
      
    self._scheduling_thread.set_timeout(sleep_time)

  # This computes the time until the next full report should be sent. If a full
  # report needs to be sent now, then this will return a negative number.
  #
  # This is called by the scheduling thread.
  def _time_until_next_full_report(self):
    if self._last_full_report_scheduled is None:
      return -1
    return self._last_full_report_scheduled + \
           self._reporting_interval_seconds - \
           time.time()

  # This call checks in on any new monitors that we have to see if they are
  # ready to report. It also checks if any new monitors are expired. This
  # returns a tuple of the number of monitors that are ready and the number
  # of new monitors that remain after the mutations this call has made.
  #
  # This is called by the scheduling thread.
  def _update_new_monitors(self):
    self._new_monitors_lock.acquire()
 
    monitors_to_remove = []
    for monitor, expiration in self._new_monitors.iteritems():
      if monitor.is_ready_to_report():
        LOG.info("Monitor ready to report: %s", (monitor,))
        monitors_to_remove.append(monitor)
        self._new_monitors_ready.append(monitor)
      elif time.time() > expiration:
        LOG.info("Monitor expired: %s", (monitor,))
        monitors_to_remove.append(monitor)

    for monitor in monitors_to_remove:
      del self._new_monitors[monitor]

    monitors_ready = len(self._new_monitors_ready)
    monitors_remaining = len(self._new_monitors)

    self._new_monitors_lock.release()

    return monitors_ready, monitors_remaining

  # Called to indicate that a new monitor has been created. This stores the new
  # monitor in the new_monitors map and wakes the scheduling thread.
  def new_monitor(self, monitor):
    LOG.info("New monitor: %s", (monitor,))
    self._new_monitors_lock.acquire()
    self._new_monitors[monitor] = time.time() + 60.0
    self._new_monitors_lock.release()
    self._scheduling_thread.wake(CountDownLatch(1))
