# Copyright (c) 2011 Cloudera, Inc. All rights reserved.

import logging
import time

from cmf.monitor import schema
from cmf.monitor.abstract_mapreduce_monitor import AbstractMapReduceMonitor
from cmf.monitor.base import DebugStatusReport
from cmf.monitor.jobtracker.json_parser import JsonParser
from cmf.monitor.json_metrics_utils import ParseError, add_jvm_metrics
from cmf.monitor.proc_metrics_utils import add_proc_metrics
from cmf.monitor.role_update import RoleUpdate
from cmf.monitor.service_update import ServiceUpdate

LOG = logging.getLogger('JobTrackerMonitor')

_ROLETYPE = 'JOBTRACKER'
_SERVICETYPE = 'MAPREDUCE'

# Section and key names in cloudera-monitor.properties
# Since CM 3.7
_SECTION = 'jobtracker'
_SERVICE_NAME = 'service_name'
_ROLE_NAME = 'role_name'
_HOST = 'host'
_MAPRED_JOB_TRACKER_HTTP_PORT = 'mapred_job_tracker_http_port'
_COLLECT_INTERVAL = 'collect_interval'
# Added in CM 4.1
_LOG_DIR = 'log_dir'

class JobTrackerMonitor(AbstractMapReduceMonitor):

  def __init__(self, daemon, registry):
    AbstractMapReduceMonitor.__init__(self,
                                      daemon,
                                      registry,
                                      _SECTION,
                                      _MAPRED_JOB_TRACKER_HTTP_PORT)
    self._last_conf_path = None
    self._metrics_tuple = (None, None, None)
    self._last_role_reported_timestamp = None
    self._last_service_reported_timestamp = None
    self._last_collection_ended = None
    self._last_collection_length = None
    self._role_dir_monitors = [ ]

    # Config data
    self._name = None
    self._service_name = None
    self._log_dir = None

  def __str__(self):
    return "JobTrackerMonitor for %s" % (self._name,)

  def _reload_config(self, confparse):
    """Returns True/False for success"""
    try:
      # Extraction
      name = confparse.get(_SECTION, _ROLE_NAME)
      service = confparse.get(_SECTION, _SERVICE_NAME)
      host = confparse.get(_SECTION, _HOST)
      log_dir = confparse.get_with_default(_SECTION, _LOG_DIR, None)
      port = confparse.getint(_SECTION, _MAPRED_JOB_TRACKER_HTTP_PORT)
    except ValueError, ex:
      self.record_error(ex)
      LOG.error("Value error in %s: %s" % (confparse.path, ex))
      return False

    metrics_url = self._compute_metrics_url(host, port)
    if metrics_url is None:
      return False

    # Assign
    self._name = name
    self._service_name = service
    self._log_dir = log_dir
    self._metrics_url = metrics_url

    # Make monitors for the log directory and the data directories.
    self._role_dir_monitors = [ ]
    if self._log_dir is not None:
      self.create_dir_monitor(self._log_dir, self._role_dir_monitors)

    return True

  def collect(self):
    """Overrides AbstractMonitor. Collect JobTracker metrics. Invoked 
       asynchronously."""
    start = time.time()

    # Reload config if necessary
    confparse = self.conf_parser
    if self._last_conf_path != confparse.path:
      if not self._reload_config(confparse):
        return
      self._last_conf_path = confparse.path

    # Collect directory info
    for mon in self._role_dir_monitors:
      mon.collect()

    # If our pid value is None, it means that the associated process is not
    # running. If this is the case, we skip metric collection.
    if self._pid is None:
      self._metrics_tuple = (None, None, None)
      return

    LOG.debug("Fetching JobTracker metrics at '%s'" % (self._metrics_url,))
    self._metrics_tuple = self.collect_metrics_from_url(self._metrics_url)

    self._last_collection_ended = time.time()
    self._last_collection_length = self._last_collection_ended - start

  def parse_metrics(self, json):
    return JsonParser(json)

  def get_role_update(self):
    """Returns a RoleUpdate to be sent to the firehose"""
    update = RoleUpdate(name=self._name,
                        roletype=_ROLETYPE,
                        service_name=self._service_name)

    if self.add_metrics_from_webserver(self._metrics_tuple,
                                       self._last_reported_timestamp,
                                       update):
      self._last_reported_timestamp = update.timestamp

    add_proc_metrics(update, self._pid, LOG)

    for mon in self._role_dir_monitors:
      update.add_dir_update(mon.get_role_dir_update())

    return update

  def get_service_update(self):
    (metrics, timestamp, duration) = self._metrics_tuple
    if metrics is None:
      return None

    # We don't want to double report the same set of metrics.
    if timestamp == self._last_service_reported_timestamp:
      return None
    self._last_service_reported_timestamp = timestamp

    mr_update = ServiceUpdate(name=self._service_name,
                              service_type=_SERVICETYPE,
                              timestamp=timestamp)
    self._add_service_metrics_update(mr_update, metrics)
    return mr_update

  def _add_service_metrics_update(self, update, metrics):
    # The metrics page is sometimes empty if the NN happens to
    # be flushing the metrics when we fetch it.
    if len(metrics.jobtracker) == 0:
      return

    update.add_metric(schema.MAP_SLOTS,
                      metrics.jobtracker.map_slots)
    update.add_metric(schema.MAPS_RUNNING,
                      metrics.jobtracker.occupied_map_slots)
    update.add_metric(schema.REDUCE_SLOTS,
                      metrics.jobtracker.reduce_slots)
    update.add_metric(schema.REDUCES_RUNNING,
                      metrics.jobtracker.occupied_reduce_slots)
    update.add_metric(schema.MR_BLACKLISTED_TT,
                      metrics.jobtracker.trackers_blacklisted)
    update.add_metric(schema.MR_JOBS_COMPLETED,
                      metrics.jobtracker.jobs_completed)
    update.add_metric(schema.MR_JOBS_FAILED,
                      metrics.jobtracker.jobs_failed)
    update.add_metric(schema.MR_JOBS_KILLED,
                      metrics.jobtracker.jobs_killed)
    update.add_metric(schema.MR_JOBS_PREPARING,
                      metrics.jobtracker.jobs_preparing)
    update.add_metric(schema.MR_JOBS_RUNNING,
                      metrics.jobtracker.jobs_running)
    update.add_metric(schema.MR_JOBS_SUBMITTED,
                      metrics.jobtracker.jobs_submitted)
    update.add_metric(schema.NUM_FAILED_MAPS,
                      metrics.jobtracker.maps_failed)
    update.add_metric(schema.NUM_FAILED_REDUCES,
                      metrics.jobtracker.reduces_failed)
    # Work-around for MAPREDUCE-1238
    update.add_metric(schema.MR_MAPS_WAITING,
                      max(0, metrics.jobtracker.waiting_maps))
    update.add_metric(schema.MR_REDUCES_WAITING,
                      max(0, metrics.jobtracker.waiting_reduces))

  def add_webserver_metrics_to_role_update(self, update, metrics):
    # The metrics page is sometimes empty if the JobTracker happens to
    # be flushing the metrics when we fetch it.
    if len(metrics.jobtracker) == 0:
      return

    add_jvm_metrics(update, metrics)

  def get_debug_status_report(self):
    return DebugStatusReport(config_items=self.conf_parser.items(_SECTION),
                             errors=(self.last_error,),
                             last_collection_ended=self._last_collection_ended,
                             last_collection_length=self._last_collection_length)

  def get_subject_type(self):
    return "JOBTRACKER"
