// (c) Copyright 2011-2012 Cloudera, Inc. All rights reserved.
define([
			 "cloudera/Util",
			 "cloudera/common/DateUtil",
			 "cloudera/common/Humanize",
			 "cloudera/common/I18n",
			 "cloudera/chart/AnyStockTimeBrowser",
			 "cloudera/chart/TimeRange",
			 "cloudera/chart/TimeBrowserState",
			 "cloudera/chart/TimeBrowserMarker",
			 "cloudera/chart/TimeBrowserURLUpdater",
			 "cloudera/chart/TimeBrowserAutoUpdater",
			 "cloudera/chart/TimeBrowserServerUpdater",
			 "cloudera/chart/TimeLabelUpdater",
			 "underscore"
], function(Util, DateUtil, Humanize, I18n, AnyStockTimeBrowser,
						TimeRange, TimeBrowserState, TimeBrowserMarker,
						TimeBrowserURLUpdater, TimeBrowserAutoUpdater,
						TimeBrowserServerUpdater, TimeLabelUpdater, _) {
	// Responsibilities:
	// -----------------
	// * Controls the interaction with the toolbar.
	// * Controls the interaction with the the custom date dialog.
	// * Delegates all other logic to the time browser.
	// * Listens for two special page events: switchToCurrent, expandTimeRangeSelection.
	//
	// Use Cases:
	// ----------
	// * The Time Browser:
	// * stores all its state in a TimeBrowserState object.
	// * instantiates the various helpers classes.
	// * the various helper objects are invoked by publish/subscribe mechanism.
	// * delegates calls to AnyStockTimeBrowser for any background line chart.
	// * delegates calls to TimeBrowserMarker for any marker movement.
	// * delegates calls to TimeBrowserURL for URL changes.
	//
	// * The Toolbar:
	// * The Fast forward icon should switch the control to current mode. If it is already
	//   in the current mode, it should be disabled.
	// * The zoom out icon should expand the total range out by a factor of 2.
	// *     In the current mode, the new right hand edge should touch the selected range, but
	//       the left hand edge should go back by a factor of 2.
	// * The zoom in icon should narrow the total range by a factor of 2, until
	//       the total range is equal to the selected time range.
	//
	// * The Custom Date Dialog:
	// * The custom date icon should bring up a popup. Clicking on any of the preset icons
	//   should change the selection to the specified range, and make the total range equal to 3 x of the selection,
	//   and hide the dialog.
	//
	// * Clicking the From/To fields involves the date/time picker,
	//   selection is made same as a preset icon after clicking OK.
	return function(options, id) {
		var self = this;

		var MS_IN_ONE_MINUTE = 60000;
		var MS_IN_HALF_HOUR = 30 * MS_IN_ONE_MINUTE;

		this.showRange = options.showRange;
		this.showMarker = options.showMarker;
		this.state = new TimeBrowserState(options);

		var options2 = $.extend({state: this.state}, options);
		this.lineChart = new AnyStockTimeBrowser(options2, id);
		this.marker = new TimeBrowserMarker(options, this.state);
		this.labelUpdater = new TimeLabelUpdater();
		this.urlUpdater = new TimeBrowserURLUpdater(options, this.state);
		this.serverUpdater = new TimeBrowserServerUpdater(options, this.state);
		if (options.showRange || options.showMarker) {
			this.autoUpdater = new TimeBrowserAutoUpdater(options, this.state);
			this.autoUpdater.start();
		}

		var toSelector = function(key) {
			return "#" + id + "_" + key;
		};

		// @return the custom date dialog element.
		var _getCustomDateDialog = function() {
			return $(toSelector("customDateDialog"));
		};

		// @return the start date datepicker element inside the custom date dialog.
		var _getStartDateElem = function() {
			return $(toSelector("customDateDialog_startDate"));
		};

		// @return the end date datepicker element inside the custom date dialog.
		var _getEndDateElem = function() {
			return $(toSelector("customDateDialog_endDate"));
		};

		// @return the custom date selector button.
		var _getDateSelectorElem = function() {
			return $(toSelector("dateSelector"));
		};

		// @return the error element in the custom date selector dialog.
		var _getDateSelectorErrorElem = function() {
			return $(toSelector("customDateDialog_error"));
		};

		// @return the zoom in button.
		var _getZoomInElem = function() {
			return $(toSelector("zoomIn"));
		};

		// @return the zoom out button.
		var _getZoomOutElem = function() {
			return $(toSelector("zoomOut"));
		};

		// @return the OK button inside the custom date dialog.
		var _getCustomDateDialogOKElem = function() {
			return $(toSelector("customDateDialog_ok"));
		};

		// For button links that should not proceed to.
		// Some calls go all the way to the ActionScript layer, and
		// exceptions are thrown. When that happens,
		// I don't want the URL hash to change to something inconsistent.
		var ignoreAnchorEvent = function(event) {
			if (event) {
				event.preventDefault();
			}
		};

		// Handles the clicking event of the Zoom In button.
		var onZoomInClicked = function(event) {
			ignoreAnchorEvent(event);
			self.state.zoomIn();
		};

		// Handles the clicking event of the Zoom Out button.
		var onZoomOutClicked = function(event) {
			ignoreAnchorEvent(event);
			self.state.zoomOut();
		};

		// Handles the window resize event.
		var onWindowResized = function(event) {
			self.marker.onWindowResized();
		};

		// @return the date rounded by the nearest interval.
		var toNearest = function(date, interval) {
			var time = date.getTime();
			return new Date(time - time % interval);
		};

		// @return the selected time range from the custom date dialog.
		var getCustomTimeRange = function() {
			// date in server timezone.
			var serverTimeRange = self.getCustomServerTimeRange();

			// date in the local timezone.
			var startDate = Humanize.fromServerDate(serverTimeRange.startDate);
			var endDate = Humanize.fromServerDate(serverTimeRange.endDate);

			return new TimeRange(startDate, endDate);
		};

		// Sets the time range in the custom date dialog.
		var setCustomTimeRange = function(timeRange) {
			// display using server timezone.
			var serverStartDate = Humanize.toServerDate(timeRange.startDate);
			var serverEndDate = Humanize.toServerDate(timeRange.endDate);

			self.setCustomServerTimeRange(new TimeRange(serverStartDate, serverEndDate));
		};

		// Handles the clicking event of the Custom Date Selector button.
		var onDateSelectorClicked = function(event) {
			ignoreAnchorEvent(event);

			var pos = $(this).offset();
			var iconWidth = $(this).width();
			var iconHeight = $(this).height();
			var dlgWidth = 700;
			var dlgHeight = 150;
			var dlgOptions = {
				modal: true,
				resizable: false,
				width: dlgWidth,
				height: dlgHeight,
				title: I18n.t("ui.customDate"),
				position: [pos.left - dlgWidth + iconWidth, pos.top + iconHeight + 10]
			};
			self.openCustomDateDialog(dlgOptions);
			var timeRange = new TimeRange(self.state.firstVisibleDate, self.state.lastVisibleDate);
			setCustomTimeRange(timeRange);
		};

		var showError = function(message) {
			_getDateSelectorErrorElem().removeClass("hidden").html(message);
		};

		var hideError = function() {
			_getDateSelectorErrorElem().addClass("hidden").html("");
		};

		// Handles the click event of the OK button in the date selector dialog.
		var onCustomOK = function(event) {
			ignoreAnchorEvent(event);

			var timeRange = getCustomTimeRange();

			if (timeRange.endDate.getTime() > Util.getServerNow().getTime()) {
				showError(I18n.t("ui.timeControl.validator.noToFutureTime"));
			} else if (timeRange.startDate.getTime() <= timeRange.endDate.getTime()) {
				if (DateUtil.delta(timeRange.startDate, timeRange.endDate) < MS_IN_ONE_MINUTE) {
					timeRange.endDate = new Date(timeRange.startDate.getTime() + MS_IN_ONE_MINUTE);
				}
				// if we are in current mode, dragging the range means we are off the current mode.
				// if we are not in current mode, dragging the range should not make us go into current mode.
				if (self.state.getCurrentMode()) {
					self.state.setCurrentMode(self.state.isDateCurrent(timeRange.endDate));
				}
				self.state.setSelectedRange(timeRange);
				self.state.moveMarkerIntoSelectedRange();
				self.closeCustomDateDialog();
				hideError();
			} else {
				showError(I18n.t("ui.timeControl.validator.fromLessThanTo"));
			}
		};

		// Handles the click event of a PresetRange button in the date selector dialog.
		var onPresetRangeClicked = function (minutes, event) {
			ignoreAnchorEvent(event);
			self.state.moveToNow(minutes * MS_IN_ONE_MINUTE);
		};

		// Handles the switch to current event.
		var onSwitchToCurrent = function() {
			self.state.moveToNow();
		};

		// Handles the expand time range event.
		var onExpandTimeRangeSelection = function() {
			self.state.expandRange();
		};

		// Returns true if user can interact with the range selection.
		var isRangeInteractive = function() {
			return (options.mode === "INTERACTIVE" && options.showRange);
		};

		var bindClickEvents = function() {
			if (isRangeInteractive()) {
				_getDateSelectorElem().click(onDateSelectorClicked);

				$(toSelector("30m")).click(function(e) { onPresetRangeClicked(30, e); });
				$(toSelector("1h")).click(function(e) { onPresetRangeClicked(60, e); });
				$(toSelector("2h")).click(function(e) { onPresetRangeClicked(120, e); });
				$(toSelector("6h")).click (function(e) { onPresetRangeClicked(360 ,e); });
				$(toSelector("12h")).click(function(e) { onPresetRangeClicked(720, e); });
				$(toSelector("24h")).click(function(e) { onPresetRangeClicked(1440, e); });

				self.initializeCustomDateRange();
				_getCustomDateDialogOKElem().click(onCustomOK);
			}

			_getZoomInElem().click(onZoomInClicked);
			_getZoomOutElem().click(onZoomOutClicked);
		};

		self.openCustomDateDialog = function(dlgOptions) {
			_getCustomDateDialog().dialog(dlgOptions);
		};

		self.closeCustomDateDialog = function() {
			_getCustomDateDialog().dialog('close');
		};

		self.initializeCustomDateRange = function() {
			_getStartDateElem().datetimepicker().datetimepicker('hide');
			_getEndDateElem().datetimepicker().datetimepicker('hide');
		};

		self.getCustomServerTimeRange = function() {
			var serverStartDate = _getStartDateElem().datetimepicker('getDate');
			var serverEndDate = _getEndDateElem().datetimepicker('getDate');
			return new TimeRange(serverStartDate, serverEndDate);
		};

		// Sets the time range in the custom date dialog.
		self.setCustomServerTimeRange = function(timeRange) {
			_getStartDateElem().datetimepicker('setDate', toNearest(timeRange.startDate, MS_IN_HALF_HOUR));
			_getEndDateElem().datetimepicker('setDate', toNearest(timeRange.endDate, MS_IN_HALF_HOUR));
		};

		bindClickEvents();

		// Do this once on startup.
		var handle1 = jQuery.subscribe("switchToCurrent", onSwitchToCurrent);
		var handle2 = jQuery.subscribe("expandTimeRangeSelection", onExpandTimeRangeSelection);
		self.subscriptionHandles = [handle1, handle2];

		$(window).resize(Util.throttle(onWindowResized, 1000));

		// Uncomment this for debugging purposes.
		var formatTimeString = function(date) {
			return Humanize.humanizeDateTimeMedium(date);
		};

		var updateDebug = function() {
			var tbState = self.state;
			$("#cmsTimeControl_first").html(formatTimeString(tbState.firstDate));
			$("#cmsTimeControl_start").html(formatTimeString(tbState.firstVisibleDate));
			$("#cmsTimeControl_markerLabel").html(formatTimeString(tbState.markerDate));
			$("#cmsTimeControl_end").html(formatTimeString(tbState.lastVisibleDate));
			$("#cmsTimeControl_last").html(formatTimeString(tbState.lastDate));
			$("#cmsTimeControl_current").html(tbState.getCurrentMode());
		};
		// Uncomment this for debugging
		// setInterval(updateDebug, 2000);

		self.unsubscribe = function() {
			Util.unsubscribe(self);
			if (self.marker) {
				self.marker.unsubscribe();
			}
		};
	};
});
